from sqjobs.utils import get_jobs_from_module

import logging
logger = logging.getLogger('sqjobs.contrib.django.utils')


def get_apps_names():
    """
    copied from django-extensions compatibility sheam
    """
    try:
        # django >= 1.7, to support AppConfig
        from django.apps import apps
        return [app.name for app in apps.get_app_configs()]
    except ImportError:
        from django.db import models
        return [app.__name__[:-7] for app in models.get_apps()]


def register_all_jobs(worker):
    """
    Register all the jobs in a worker
    """
    jobs = get_all_jobs()

    for job in jobs:
        worker.register_job(job)

    return jobs


def get_all_jobs():
    """
    Get all the jobs of the django INSTALLED_APPS
    """
    jobs = []

    for app_name in get_apps_names():
        try:
            module = app_name + '.jobs'
            jobs.extend(get_jobs_from_module(module))
        except ImportError:
            pass

    return jobs



import sys
import traceback

from .exceptions import RetryException

import logging
logger = logging.getLogger('sqjobs.worker')


class Worker(object):
    DEFAULT_TIMEOUT = 20  # seconds

    def __init__(self, broker, queue_name, timeout=None):
        self.broker = broker
        self.queue_name = queue_name
        self.timeout = timeout or self.DEFAULT_TIMEOUT
        self.registered_jobs = {}
        self.exception_handlers = []

    def __repr__(self):
        return 'Worker({connector})'.format(
            connector=type(self.broker.connector).__name__
        )

    def register_job(self, job_class):
        name = job_class._task_name()

        if job_class.abstract:
            logger.info('Job %s is abstract, ignoring it...', name)
            return

        if name in self.registered_jobs:
            logger.warning('Job %s already registered, overwriting it...', name)

        logger.info('Registering new job: %s', name)

        self.registered_jobs[name] = job_class

    def append_exception_handler(self, handler):
        self.exception_handlers.append(handler)

    def run(self):
        logger.info('Running worker, %d jobs registered...', len(self.registered_jobs))

        for payload in self.broker.jobs(self.queue_name, self.timeout):
            try:
                job_class = self.registered_jobs.get(payload['name'])

                if not job_class:
                    logger.error('Unregistered task: %s', payload['name'])
                    continue

                job, args, kwargs = self.broker.unserialize_job(job_class, payload)
                self._execute_job(job, args, kwargs)
            except:
                logger.exception('Error executing job')

    def _set_custom_retry_time_if_needed(self, job):
        if job.next_retry_time() is None:  # Use default value of the queue
            return

        self.broker.set_retry_time(job, job.next_retry_time())

    def _execute_job(self, job, args, kwargs):
        try:
            job.execute(*args, **kwargs)
            self.broker.delete_job(job)
        except RetryException:
            job.on_retry()
            self._set_custom_retry_time_if_needed(job)
            return
        except:
            job.on_failure()
            self._handle_exception(job, args, kwargs, *sys.exc_info())
            self._set_custom_retry_time_if_needed(job)
            return

        job.on_success()

    def _handle_exception(self, job, args, kwargs, *exc_info):
        exception_message = ''.join(
            traceback.format_exception_only(*exc_info[:2]) +
            traceback.format_exception(*exc_info)
        )

        logger.error(exception_message, exc_info=True, extra={
            'job_queue_name': job.queue_name,
            'job_id': job.id,
            'job_name': job.name,
            'job_args': args,
            'job_kwargs': kwargs,
        })

        for handler in reversed(self.exception_handlers):
            logger.debug('Executing exception handler %s', handler)
            handler(job, args, kwargs, *exc_info)



import json
from abc import ABCMeta
from six import add_metaclass
from datetime import datetime

from django.db import IntegrityError

from sqjobs import Job
from sqjobs.contrib.django.djsqjobs.models import JobStatus


@add_metaclass(ABCMeta)
class ResultJob(Job):
    """
    A job that stores the result of the execution in the DB
    """

    def __init__(self):
        self.properly_setup = False
        self.repeated_task = False

    def execute(self, *args, **kwargs):
        self.pre_run(*args, **kwargs)

        if not self.repeated_task:
            self.result = self.run(*args, **kwargs)
            self.post_run(*args, **kwargs)

    def pre_run(self, *args, **kwargs):
        super(ResultJob, self).pre_run(*args, **kwargs)

        # Take as id the left part from the hash symbol
        task_id = self.id.split('#')[0]
        try:
            self.job_status = JobStatus(
                job_id=task_id,
                job_name=self.__repr__()
            )
            self.job_status.save(force_insert=True)
            self.properly_setup = True
        except IntegrityError:
            self.job_status = JobStatus.objects.get(job_id=task_id)
            if self.job_status.status == JobStatus.FAILURE:
                self.job_status.status = JobStatus.PENDING
                self.job_status.save()
                self.properly_setup = True
            else:
                self.repeated_task = True

    def post_run(self, *args, **kwargs):
        self.job_status.date_done = datetime.now()
        self.job_status.result = json.dumps(self.result)
        self.job_status.save(force_update=True)
        super(ResultJob, self).post_run(*args, **kwargs)

    def on_success(self, *args, **kwargs):
        if not self.repeated_task:
            self.job_status.status = JobStatus.SUCCESS
            self.job_status.save(force_update=True)
            super(ResultJob, self).on_success(*args, **kwargs)

    def on_failure(self):
        if not self.repeated_task:
            if self.properly_setup:
                self.job_status.status = JobStatus.FAILURE
                self.job_status.save(force_update=True)
            super(ResultJob, self).on_failure()



from abc import ABCMeta, abstractmethod
from six import add_metaclass


@add_metaclass(ABCMeta)
class Connector(object):
    """
    Manages a single connection to the broker
    """

    @abstractmethod
    def enqueue(self, queue_name, payload):
        """
        Sends a new message to a queue

        :param queue_name: the name of the queue
        :param payload: the payload to send inside the message
        """
        raise NotImplementedError

    @abstractmethod
    def dequeue(self, queue_name, wait_time=20):
        """
        Receives the next message from a queue

        :param queue_name: the queue name
        :param wait_time: how much time to wait until a new message is retrieved (long polling).
         If set to zero, connection will return inmediately if no messages exist.
        """
        raise NotImplementedError

    @abstractmethod
    def delete(self, queue_name, message_id):
        """
        Deletes a message from a queue

        :param queue_name: the name of the queue
        :param message_id: the message id
        """
        raise NotImplementedError

    @abstractmethod
    def set_retry_time(self, queue_name, message_id, delay):
        """
        Tell the connector to retry a job

        :param queue_name: the name of the queue
        :param message_id: the message id
        :param delay: delay (in seconds) when the next retry will be attempted
        """
        raise NotImplementedError

    @abstractmethod
    def serialize_job(self, job_class, job_id, args, kwargs):
        """
        Serialize a job into a string to be sent to the connector

        :param job_class: Python class of the payload job
        :param job_id: the ID of the job
        :param args: arguments of the job
        :param kwargs: keyword arguments of the job
        """
        raise NotImplementedError

    @abstractmethod
    def unserialize_job(self, job_class, payload):
        """
        Build a job given a payload returned from the connector

        :param job_class: Python class of the payload job
        :param payload: Python dict with the job arguments
        """
        raise NotImplementedError



from .base import Broker
from ..job import Job, JobResult


class Standard(Broker):
    """
    Standard broker to execute jobs in an asynchronous way
    """

    def __init__(self, connector):
        self.connector = connector

    def __repr__(self):
        return 'Broker({connector})'.format(
            connector=type(self.connector).__name__
        )

    def add_job(self, job_class, *args, **kwargs):
        job_name = job_class._task_name()
        queue_name = kwargs.pop('queue_name', job_class.default_queue_name)

        return self.add_job_by_name(job_name, queue_name, *args, **kwargs)

    def add_job_by_name(self, job_name, queue_name, *args, **kwargs):
        job_id = self.gen_job_id()

        payload = self.serialize_job(job_name, job_id, args, kwargs)
        self.connector.enqueue(queue_name, payload)

        result = JobResult()
        result.job_id = job_id

        return result

    def delete_job(self, job):
        self.connector.delete(job.queue_name, job.broker_id)

    def retry(self, job):
        self.set_retry_time(job, 0)

    def set_retry_time(self, job, retry_time):
        self.connector.set_retry_time(job.queue_name, job.broker_id, retry_time)

    def jobs(self, queue_name, timeout=20):
        while True:
            payload = self.connector.dequeue(queue_name, wait_time=timeout)
            if payload or not timeout:
                yield payload



import pytest

from ..connectors.dummy import Dummy as DummyConnector
from ..worker import Worker
from ..brokers.standard import Standard as StandardBroker
from .fixtures import Adder, FakeAdder, AbstractAdder, RetryJob, ExceptionJob


class TestWorker(object):

    @property
    def connector(self):
        return DummyConnector()

    @property
    def broker(self):
        return StandardBroker(self.connector)

    def test_worker_repr(self):
        worker = Worker(self.broker, 'default')
        assert repr(worker) == 'Worker(Dummy)'

    def test_register_job(self):
        worker = Worker(self.broker, 'default')
        worker.register_job(Adder)

        assert len(worker.registered_jobs) == 1
        assert worker.registered_jobs[Adder.name] == Adder

    def test_try_register_abstract_job(self):
        worker = Worker(self.broker, 'default')
        worker.register_job(AbstractAdder)

        assert len(worker.registered_jobs) == 0

    def test_register_same_job_twice(self):
        worker = Worker(self.broker, 'default')
        worker.register_job(Adder)
        worker.register_job(Adder)

        assert len(worker.registered_jobs) == 1
        assert worker.registered_jobs[Adder.name] == Adder

    def test_overwrite_job_when_register_with_same_name(self):
        worker = Worker(self.broker, 'default')
        worker.register_job(Adder)
        worker.register_job(FakeAdder)

        assert len(worker.registered_jobs) == 1
        assert worker.registered_jobs[FakeAdder.name] == FakeAdder



import pytest

from .fixtures import Adder, Divider
from ..brokers.standard import Standard as StandardBroker
from ..brokers.eager import Eager as EagerBroker
from ..brokers.base import Broker as BaseBroker
from ..connectors.dummy import Dummy as DummyConnector


class TestBrokerInterface(object):

    def test_default_broker_interface(self):
        with pytest.raises(NotImplementedError):
            BaseBroker.add_job(StandardBroker(None), Adder, 'demo', {})


class TestStandardBroker(object):

    @property
    def connector(self):
        return DummyConnector()

    def test_broker_repr(self):
        broker = StandardBroker(self.connector)
        assert repr(broker) == 'Broker(Dummy)'

    def test_add_job_to_broker(self):
        broker = StandardBroker(self.connector)
        broker.add_job(Adder, 2, 3)

        queues = list(broker.connector.jobs.keys())
        assert len(queues) == 1
        assert queues[0] == 'sqjobs'
        assert broker.connector.num_jobs == 1

        messages = broker.connector.jobs['sqjobs']
        assert len(messages) == 1

    def test_add_job_by_name_to_broker(self):
        broker = StandardBroker(self.connector)
        broker.add_job_by_name('Adder', 'sqjobs', 2, 3)

        queues = list(broker.connector.jobs.keys())
        assert len(queues) == 1
        assert queues[0] == 'sqjobs'
        assert broker.connector.num_jobs == 1

        messages = broker.connector.jobs['sqjobs']
        assert len(messages) == 1

    def test_right_payload_args_when_job_is_added(self):
        broker = StandardBroker(self.connector)
        broker.add_job(Adder, 2, 3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (2, 3), 'kwargs': {}, 'name': 'adder'}

    def test_right_payload_args_when_job_is_added_by_name(self):
        broker = StandardBroker(self.connector)
        broker.add_job_by_name('adder', 'sqjobs', 2, 3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (2, 3), 'kwargs': {}, 'name': 'adder'}

    def test_right_payload_kwargs_when_job_is_added(self):
        broker = StandardBroker(self.connector)
        broker.add_job(Adder, num2=2, num1=3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (), 'kwargs': {'num1': 3, 'num2': 2}, 'name': 'adder'}

    def test_right_payload_kwargs_when_job_is_added_by_name(self):
        broker = StandardBroker(self.connector)
        broker.add_job_by_name('adder', 'sqjobs', num2=2, num1=3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (), 'kwargs': {'num1': 3, 'num2': 2}, 'name': 'adder'}

    def test_right_payload_both_when_job_is_added(self):
        broker = StandardBroker(self.connector)
        broker.add_job(Adder, 2, num2=3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (2,), 'kwargs': {'num2': 3}, 'name': 'adder'}

    def test_right_payload_both_when_job_is_added_by_name(self):
        broker = StandardBroker(self.connector)
        broker.add_job_by_name('adder', 'sqjobs', 2, num2=3)

        message = broker.connector.jobs['sqjobs'][0]
        del message['id']  # Check that exists, but we don't care about the value

        assert message == {'args': (2,), 'kwargs': {'num2': 3}, 'name': 'adder'}

    def test_multiple_jobs_are_stored_correctly_by_the_broker(self):
        broker = StandardBroker(self.connector)
        job_ids, jobs = [], []

        job_ids.append(broker.add_job(Adder, 1, 1))
        job_ids.append(broker.add_job(Adder, 2, 2))

        gen = broker.jobs('sqjobs')

        for _ in range(2):
            jobs.append(next(gen))

        assert len(jobs) == 2
        assert jobs[0] == {'id': job_ids[1].job_id, 'args': (2, 2), 'kwargs': {}, 'name': 'adder'}
        assert jobs[1] == {'id': job_ids[0].job_id, 'args': (1, 1), 'kwargs': {}, 'name': 'adder'}


class TestEagerBroker(object):

    def test_broker_repr(self):
        broker = EagerBroker()
        assert repr(broker) == 'Broker(Eager)'

    def test_execute_job_eager_mode(self):
        broker = EagerBroker()
        result = broker.add_job(Adder, 2, 3)
        assert result.result == 5

    def test_eager_failure(self):
        broker = EagerBroker()

        with pytest.raises(ZeroDivisionError):
            assert broker.add_job(Divider, 2, 0)



from time import time, sleep

from .standard import Standard


class MultiQueue(Standard):
    """
    Broker to execute jobs in an asynchronous way from multiple queues.
    Execute a job from one queue and go to the next one, if the queue is empty
    add a timestamp to wait `polling_interval` seconds before trying again.
    """
    polling_interval = 20  # seconds

    def jobs(self, queue_names, timeout=0):
        """
        * queue_names: List of queue names from where to read.
        * timeout: Not used by this implementation. It's kept for compatibility.
        """
        queue_last_check = {queue: 0 for queue in queue_names}

        # loop through the queue list forever
        while True:
            for queue in queue_names:
                next_check = queue_last_check[queue] + self.polling_interval

                # if it's been less than `polling_time` since last check, continue
                if next_check > int(time()):
                    continue

                # wait_time needs to be 0 or it will block in the first queue
                payload = self.connector.dequeue(queue, wait_time=0)
                if payload:
                    queue_last_check[queue] = 0
                    yield payload
                else:
                    queue_last_check[queue] = int(time())

            # only sleep if all queues are empty
            if all(value != 0 for value in queue_last_check.values()):
                sleep(self.polling_interval)



from django.conf import settings

from sqjobs import create_sqs_broker, create_sqs_worker


def get_broker():
    return create_sqs_broker(
        access_key=settings.SQJOBS_SQS_ACCESS_KEY,
        secret_key=settings.SQJOBS_SQS_SECRET_KEY,
        region_name=settings.SQJOBS_SQS_REGION_NAME,
        endpoint_url=getattr(settings, 'SQJOBS_SQS_ENDPOINT_URL', None),
    )


def get_worker(queue_name):
    return create_sqs_worker(
        queue_name=queue_name,
        access_key=settings.SQJOBS_SQS_ACCESS_KEY,
        secret_key=settings.SQJOBS_SQS_SECRET_KEY,
        region_name=settings.SQJOBS_SQS_REGION_NAME,
        endpoint_url=getattr(settings, 'SQJOBS_SQS_ENDPOINT_URL', None),
    )


def add_job(job_class, *args, **kwargs):
    get_broker().add_job(job_class, *args, **kwargs)



from __future__ import absolute_import

from django.core.management.base import BaseCommand

from sqjobs.contrib.django.djsqjobs.finders import register_all_jobs
from sqjobs.contrib.django.djsqjobs.utils import get_worker


class Command(BaseCommand):
    help = 'sqjobs commands'
    args = True

    def handle(self, *args, **options):
        if len(args) != 3 or args[0] != 'sqs' or args[1] != 'worker':
            self.help_text()
            return

        self._execute_worker(args[2])

    def _execute_worker(self, queue_name):
        worker = get_worker(queue_name)

        register_all_jobs(worker)
        worker.run()

    def help_text(self):
        self.stdout.write('Use:')
        self.stdout.write('./manage.py sqjobs sqs worker QUEUE_NAME')



from sqjobs import Job


class Adder(Job):
    name = 'adder'
    default_queue_name = 'sqjobs'
    retry_time = 10

    def run(self, num1, num2):
        return num1 + num2


class Divider(Job):
    name = 'divider'
    default_queue_name = 'math_operations'
    retry_time = 10

    def pre_run(self, num1, num2):
        self.num1 = num1 + 1

    def run(self, num1, num2):
        return self.num1 // num2

    def post_run(self, num1, num2):
        self.result = str(self.result)

    def on_success(self):
        self.status = 'OK'

    def on_failure(self):
        self.status = 'KO'

    def on_retry(self):
        self.status = 'RETRIED'

class RetryJob(Job):
    name = 'retry'
    retried_callback_called = False

    @classmethod
    def has_retried(cls):
        return cls.retried_callback_called

    @classmethod
    def retried(cls):
        cls.retried_callback_called = True

    @classmethod
    def reset_retried(cls):
        cls.retried_callback_called = True

    def __init__(self):
        super(RetryJob, self).__init__()
        RetryJob.reset_retried()

    def run(self):
        self.retry()

    def on_retry(self):
        RetryJob.retried()


class ExceptionJob(Job):
    name = 'exception'
    failed_callback_called = False
    retry_time = 10

    @classmethod
    def has_failed(cls):
        return cls.failed_callback_called

    @classmethod
    def failed(cls):
        cls.failed_callback_called = True

    @classmethod
    def reset_failed(cls):
        cls.failed_callback_called = True

    def __init__(self):
        super(ExceptionJob, self).__init__()
        ExceptionJob.reset_failed()

    def run(self):
        raise Exception("Test")

    def on_fail(self):
        ExceptionJob.retried()


class FakeAdder(Adder):
    retry_time = None


class AbstractAdder(Adder):
    abstract = True


class ComplexRetryJob(Adder):
    name = 'complex'
    retry_time = 10

    def next_retry_time(self):
        return (self.retries + 1) * self.retry_time



def create_raven_client(dsn):
    from raven import Client
    return Client(dsn)


def register_sentry(raven_client, worker):
    """
    Register sentry to send worker exceptions to it
    """

    def send_exception_to_sentry(job, args, kwargs, *exc_info):
        raven_client.captureException(exc_info=exc_info, extra={
            'job_id': job.id,
            'job_name': job.name,
            'job_args': args,
            'job_kwargs': kwargs,
            'job_retries': job.retries,
        })

    worker.append_exception_handler(send_exception_to_sentry)



import sys
import six

from .base import Broker
from ..job import JobResult


class Eager(Broker):
    """
    Broker to execute jobs in a synchronous way
    """

    def __repr__(self):
        return 'Broker(Eager)'

    def add_job(self, job_class, *args, **kwargs):
        job_id = self.gen_job_id()

        eager_job = job_class()
        eager_job.id = job_id

        try:
            result = eager_job.execute(*args, **kwargs)
            eager_job.on_success()
        except Exception as e:
            eager_job.on_failure()
            six.reraise(*sys.exc_info())

        job_result = JobResult()
        job_result.job_id = job_id
        job_result.result = result

        return job_result



from .exceptions import RetryException
from .job import Job
from .utils import create_eager_broker, create_sqs_broker, create_sqs_worker


__all__ = [
    'Job',
    'RetryException',
    'create_sqs_broker',
    'create_sqs_worker',
    'create_eager_broker'
]



class SQJobsException(Exception):
    """
    Base class for all the sqjobs exceptions
    """


class RetryException(SQJobsException):
    """
    Exception to raise when a retry is needed
    """


class QueueDoesNotExist(SQJobsException):
    """
    A queue does not exist in the broker
    """



from .base import Connector


class Dummy(Connector):
    """
    Dummy connector for testing purposes
    """

    def __init__(self):
        self.jobs = {}
        self.deleted_jobs = {}
        self.retried_jobs = {}

        self.num_jobs = 0
        self.num_deleted_jobs = 0
        self.num_retried_jobs = 0

    def enqueue(self, queue_name, payload):
        self._get_queue(queue_name).append(payload)
        self.num_jobs += 1

    def dequeue(self, queue_name, wait_time=20):
        job = None

        try:
            job = self._get_queue(queue_name).pop()
            self.num_jobs -= 1
        except IndexError:
            pass

        return job

    def delete(self, queue_name, message_id):
        self.deleted_jobs.setdefault(queue_name, []).append(message_id)
        self.num_deleted_jobs += 1

    def set_retry_time(self, queue_name, message_id, delay):
        self.retried_jobs.setdefault(queue_name, []).append((message_id, delay))
        self.num_retried_jobs += 1

    def serialize_job(self, job_name, job_id, args, kwargs):
        return {
            'id': job_id,
            'name': job_name,
            'args': args,
            'kwargs': kwargs
        }

    def unserialize_job(self, job_class, payload):
        job = job_class()

        job.id = payload['id']
        job.queue_name = payload['_metadata']['queue_name']
        job.broker_id = payload['_metadata']['id']
        job.retries = payload['_metadata']['retries']
        job.created_on = payload['_metadata']['created_on']
        args = payload['args'] or []
        kwargs = payload['kwargs'] or {}

        return job, args, kwargs

    def _get_queue(self, name):
        return self.jobs.setdefault(name, [])



import pytest

from sqjobs import Job, RetryException
from .fixtures import Adder, Divider, ComplexRetryJob


class TestJobs(object):

    def test_job_class_defaults(self):
        assert Job.default_queue_name == 'sqjobs'
        assert Job.retry_time is None
        assert Job.name is None
        assert Job.abstract is False
        assert Job._task_name() == 'Job'

        with pytest.raises(TypeError):
            Job()

        with pytest.raises(NotImplementedError):
            Job.run(Adder())

    def test_job_obj_defaults(self):
        adder = Adder()

        assert repr(adder) == 'Adder()'
        assert adder.id is None
        assert adder.retries == 0
        assert adder.created_on is None
        assert adder.queue_name is None
        assert adder.broker_id is None
        assert hasattr(adder, 'result') is False

    def test_overwritten_default_queue(self):
        assert Divider().default_queue_name == 'math_operations'

    def test_job_names(self):
        adder = Adder()
        assert adder.name == 'adder'
        assert adder._task_name() == 'adder'

    def test_run_job_directly(self):
        adder = Adder()
        assert adder.run(1, 3) == 4

    def test_execute_job_directly(self):
        adder = Adder()
        adder.execute(1, 3)
        assert adder.result == 4

    def test_whole_execute_lifecycle_job(self):
        divider = Divider()
        divider.execute(2, 1)
        assert divider.result == '3'

    def test_retry_raises_exception(self):
        adder = Adder()

        with pytest.raises(RetryException):
            adder.retry()

    def test_simple_next_retry(self):
        adder = Adder()
        assert adder.next_retry_time() == 10


class TestRetries(object):

    def test_first_complex_retry(self):
        job = ComplexRetryJob()
        assert job.retry_time == 10
        assert job.next_retry_time() == 10

    def test_second_complex_retry(self):
        job = ComplexRetryJob()
        job.retries = 1

        assert job.next_retry_time() == 20



#!/usr/bin/env python
import os
import sys

from django.conf import settings
import django


DEFAULT_SETTINGS = {
    'INSTALLED_APPS': (
        'sqjobs',
        'sqjobs.tests',
        'sqjobs.contrib.django.djsqjobs'
    ),
    'DATABASES': {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
        }
    },
    'USE_TZ': 'true',
    'SILENCED_SYSTEM_CHECKS': ["1_7.W001"],
}

if not settings.configured:
    settings.configure(**DEFAULT_SETTINGS)

if hasattr(django, 'setup'):
    django.setup()

parent = os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.abspath(__file__)
)))

sys.path.insert(0, parent)


from django.test.runner import DiscoverRunner

res = DiscoverRunner(failfast=False).run_tests([
    'sqjobs.tests'
], verbosity=1)

sys.exit(res)



#!/usr/bin/env python

"""
SQJobs - Simple Queue Jobs.

Usage:
  sqjobs <broker> worker [options] <queue_name>
  sqjobs (-h | --help)
  sqjobs --version

Options:
  --jobs=<module>               Python module where jobs are located [default: .jobs]

AWS SQS Options:
  --aws-access-key=<ak>         Access key to access SQS
  --aws-secret-key=<sk>         Secret key to access SQS
  --aws-region-name=<region>    AWS Region [default: us-west-1]

Utils:
  --sentry-dsn=<sentry_dsn>     Sentry DSN to report exceptions (raven must be installed)

Other options:
  -h --help              Show this screen.
  --version              Show version.
"""

import logging
import os
import sys

from docopt import docopt

from .contrib.sentry import create_raven_client, register_sentry
from .metadata import __version__
from .utils import create_sqs_worker, get_jobs_from_module

import logging
logger = logging.getLogger('sqjobs.cli')


def get_worker_config(broker, arguments):
    if broker != 'sqs':
        raise ValueError('Unknown broker: %s' % broker)

    config = {
        'access_key': arguments['--aws-access-key'],
        'secret_key': arguments['--aws-secret-key'],
        'region_name': arguments['--aws-region-name'],
    }

    if not config['access_key'] or not config['secret_key']:
        raise ValueError('--aws-access-key and --aws-secret-key are mandatory for SQS')

    return config


def config_logger(arguments):
    logging.basicConfig(
        format='[%(asctime)s][%(name)s] %(message)s',
        level='INFO',
        datefmt='%Y-%m-%d %H:%M:%S'
    )


def main():
    arguments = docopt(__doc__, version=__version__)
    config_logger(arguments)

    queue_name = arguments['<queue_name>']
    worker_config = get_worker_config(arguments['<broker>'], arguments)

    worker = create_sqs_worker(
        queue_name=queue_name,
        **worker_config
    )

    # Add the CWD to the python path
    sys.path.append(os.getcwd())

    for job in get_jobs_from_module(arguments['--jobs']):
        worker.register_job(job)

    logger.info('%d jobs registered', len(worker.registered_jobs))

    if arguments['--sentry-dsn']:
        raven_client = create_raven_client(arguments['--sentry-dsn'])
        register_sentry(raven_client, worker)

    worker.run()


if __name__ == '__main__':
    main()



__uname__ = 'sqjobs-tf'
__long_name__ = 'Simple Queue Jobs TF Fork'
__version__ = '0.8.7'
__author__ = 'Thinkful'
__email__ = 'engineering@thinkful.com'
__url__ = 'http://github.com/Thinkful/sqjobs'
__license__ = 'BSD'



import pytz
from datetime import datetime, timedelta

from croniter.croniter import croniter
from django.db import models
from django.utils.translation import ugettext_lazy as _


class JobStatus(models.Model):
    """
    Stores in the database the status of a job execution
    """
    PENDING = 'PENDING'
    SUCCESS = 'SUCCESS'
    FAILURE = 'FAILURE'

    statuses = (
        (PENDING, 'PENDING'),
        (SUCCESS, 'SUCCESS'),
        (FAILURE, 'FAILURE')
    )

    job_id = models.CharField(_('job id'), max_length=255, unique=True)
    job_name = models.CharField(_('job name'), max_length=255)
    status = models.CharField(_('status'), max_length=7, default=PENDING, choices=statuses)
    result = models.TextField(_('result'), blank=True)
    traceback = models.TextField(_('traceback'), blank=True)

    created_at = models.DateTimeField(_('started at'), auto_now_add=True)
    updated_at = models.DateTimeField(_('done at'), auto_now=True)

    def __repr__(self):
        return '<JobStatus ({name}): {id} status->{status}>'.format(
            name=self.job_name,
            id=self.job_id,
            status=self.status
        )


class PeriodicJob(models.Model):
    """
    A scheduled job
    """
    PROGRAMMED_DATE = 'sqjobs_programmed_date'
    DELAYED_JOB = 'sqjobs_delayed_job'

    name = models.CharField(_('name'), max_length=255)
    task = models.CharField(_('task'), max_length=255)
    args = models.TextField(_('args'), blank=True)
    kwargs = models.TextField(_('kwargs'), blank=True)

    schedule = models.CharField(_('schedule'), max_length=255)
    timezone = models.CharField(_('timezone'), max_length=63, default='UTC')

    created_at = models.DateTimeField(_('created at'), auto_now_add=True)
    next_execution = models.DateTimeField(_('next execution on'))

    enabled = models.BooleanField(_('enabled'), default=True)
    skip_delayed_jobs_next_time = models.BooleanField(_('skip jobs if delayed'), default=True)

    def save(self, *args, **kwargs):
        self.next_execution = self.get_next_utc_execution()
        super(PeriodicJob, self).save(*args, **kwargs)

    def get_next_utc_execution(self):
        tz = pytz.timezone(self.timezone)

        if not self.next_execution:
            base = tz.normalize(datetime.now(tz) - timedelta(minutes=1))
        elif self.skip_delayed_jobs_next_time:
            base = tz.normalize(datetime.now(tz))
        else:
            base = tz.normalize(self.next_execution.replace(tzinfo=pytz.utc))

        cron = croniter(self.schedule, base)

        return cron.get_next(datetime).astimezone(pytz.utc)



from sqjobs.contrib.django.djsqjobs.result_job import ResultJob
from sqjobs.contrib.django.djsqjobs.models import JobStatus

from ...fixtures import Divider


class ResultDivider(ResultJob):
    name = 'resultdivider'

    def __init__(self):
        super(ResultDivider, self).__init__()

    def pre_run(self, *args, **kwargs):
        super_setup = super(ResultDivider, self).pre_run(*args, **kwargs)
        return super_setup

    def run(self, num1, num2, *args, **kwargs):
        self.result = num1 // num2
        return self.result

    def on_failure(self, *args, **kwargs):
        super_on_failure = super(ResultDivider, self).on_failure(*args, **kwargs)
        self.err = 'ZeroDivisionError'
        return super_on_failure



import boto3
import botocore
from datetime import datetime
from pytz import timezone
import pytest
import mock

from sqjobs.exceptions import QueueDoesNotExist
from ..connectors.sqs import SQS, SQSMessage
from ..connectors.base import Connector
from ..connectors.dummy import Dummy
from .fixtures import Adder

QUEUE_NAME = 'my_queue'
SQS_SERVICE_NAME = 'sqs'
SQS_REGION_NAME = 'us-east-1'
SQS_ACCESS_KEY = 'ACCESS_KEY'
SQS_SECRET_KEY = 'SECRET_KEY'
ENDPOINT_URL = 'my_url'


class TestConnectorInterface(object):

    def test_default_connector_interface(self):
        dummy = Dummy()

        with pytest.raises(NotImplementedError):
            Connector.enqueue(dummy, 'demo', {})

        with pytest.raises(NotImplementedError):
            Connector.dequeue(dummy, 'demo', wait_time=20)

        with pytest.raises(NotImplementedError):
            Connector.delete(dummy, 'demo', 'id')

        with pytest.raises(NotImplementedError):
            Connector.set_retry_time(dummy, 'demo', 'id', 10)

        with pytest.raises(NotImplementedError):
            Connector.serialize_job(dummy, dummy, 'job_id', 'args', {})

        with pytest.raises(NotImplementedError):
            Connector.unserialize_job(dummy, dummy, 'payload')


class SQSMock(object):
    def __init__(self, raise_queue_not_found=False):
        self.raise_queue_not_found = raise_queue_not_found

    def get_queue_by_name(self, QueueName):
        if self.raise_queue_not_found:
            error_response = {'Error': {}}

            raise botocore.exceptions.ClientError(
                error_response=error_response,
                operation_name=None
            )

        return SQSQueueMock()


class SQSQueueMock(object):
    def send_message(self, MessageBody):
        pass

    def receive_messages(self, MaxNumberOfMessages, WaitTimeSeconds, AttributeNames):
        pass

    def delete_messages(self):
        pass

    def change_message_visibility_batch(self):
        pass


class SQSMessageMock(object):

    def __init__(self, receipt_handle="1"):
        self.body = SQSMessage.encode({'key': 'value'})
        self.receipt_handle = receipt_handle
        self.attributes = {
            "ApproximateReceiveCount": 0,
            "SentTimestamp": 1460186204000
        }


class TestSQSConnector(object):

    def create_sqs_connector(self):
        sqs_connector = SQS(
            access_key=SQS_ACCESS_KEY,
            secret_key=SQS_SECRET_KEY,
            region_name=SQS_REGION_NAME
        )

        return sqs_connector

    def create_sqs_connector_with_endpoint_url(self):
        sqs_connector = self.create_sqs_connector()
        sqs_connector.endpoint_url = ENDPOINT_URL
        return sqs_connector

    @mock.patch.object(boto3, 'resource')
    def test_connection_is_cached(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        sqs_connector.connection

        assert sqs_connector._cached_connection is not None

    @mock.patch.object(boto3, 'resource')
    def test_connection_created_with_aws_parameters(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        sqs_connector.connection

        sqs_mock.assert_called_with(
            aws_access_key_id=SQS_ACCESS_KEY,
            aws_secret_access_key=SQS_SECRET_KEY,
            region_name=SQS_REGION_NAME,
            service_name=SQS_SERVICE_NAME
        )

    @mock.patch.object(boto3, 'resource')
    def test_connection_created_with_endpoint_url_if_provided(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector_with_endpoint_url()

        sqs_connector.connection

        sqs_mock.assert_called_with(
            aws_access_key_id=SQS_ACCESS_KEY,
            aws_secret_access_key=SQS_SECRET_KEY,
            region_name=SQS_REGION_NAME,
            service_name=SQS_SERVICE_NAME,
            endpoint_url=ENDPOINT_URL
        )

    def test_serialize_job_returns_valid_json(self):
        sqs_connector = self.create_sqs_connector()

        expected_serializer = {
            'args': [3, 4],
            'id': 15,
            'kwargs': {'first_param': 1, 'second_param': 'two'},
            'name': 'adder'
        }

        serializer = sqs_connector.serialize_job(
            job_name='adder',
            job_id=15,
            args=[3, 4],
            kwargs={'first_param': 1, 'second_param': 'two'}
        )

        assert serializer == expected_serializer

    def test_unserialize_job_returns_job_class(self):
        sqs_connector = self.create_sqs_connector()

        payload = {
            'id': 1,
            '_metadata': {
                'created_on': datetime(2016, 4, 9, 7, 16, 44, tzinfo=timezone('UTC')),
                'id': '1',
                'retries': 0,
                'queue_name': 'my_queue'
            },
            'args': [1, 'second_arg'],
            'kwargs': {
                'first_kwarg': 1,
                'second_kwarg': '2'
            }
        }

        job, args, kwargs = sqs_connector.unserialize_job(
            job_class=Adder,
            payload=payload
        )

        assert isinstance(job, Adder) is True
        assert args == [1, 'second_arg']
        assert kwargs == {'first_kwarg': 1, 'second_kwarg': '2'}

    @mock.patch.object(boto3, 'resource')
    def test_connection_enqueue_message_goes_ok(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        with mock.patch.object(SQSQueueMock, 'send_message') as send_message_mock:
            sqs_connector.enqueue(queue_name=('%s' % QUEUE_NAME), payload="{'key': value}")
            send_message_mock.assert_called_with(
                MessageBody="Insna2V5JzogdmFsdWV9Ig=="
            )

    @mock.patch.object(boto3, 'resource')
    def test_connection_enqueue_message_fails_if_no_queue_found(self, sqs_mock):
        sqs_mock.return_value = SQSMock(raise_queue_not_found=True)
        sqs_connector = self.create_sqs_connector()

        pytest.raises(QueueDoesNotExist, sqs_connector.enqueue, queue_name=QUEUE_NAME, payload="{}")

    @mock.patch.object(boto3, 'resource')
    def test_connection_dequeue_message_goes_ok(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        expected_payload = {
            '_metadata': {
                'created_on': datetime(2016, 4, 9, 7, 16, 44, tzinfo=timezone('UTC')),
                'id': '1',
                'retries': 0,
                'queue_name': 'my_queue'
            },
            'key': 'value'
        }

        with mock.patch.object(SQSQueueMock, 'receive_messages') as receive_messages_mock:
            receive_messages_mock.return_value = [
                SQSMessageMock(receipt_handle="1"),
                SQSMessageMock(receipt_handle="2")
            ]

            payload = sqs_connector.dequeue(QUEUE_NAME)

            receive_messages_mock.assert_called_with(
                AttributeNames=['All'],
                MaxNumberOfMessages=1,
                WaitTimeSeconds=20
            )

        assert payload == expected_payload

    @mock.patch.object(boto3, 'resource')
    def test_connection_dequeue_message_fails_if_no_queue_found(self, sqs_mock):
        sqs_mock.return_value = SQSMock(raise_queue_not_found=True)
        sqs_connector = self.create_sqs_connector()

        pytest.raises(QueueDoesNotExist, sqs_connector.dequeue, queue_name=QUEUE_NAME)

    @mock.patch.object(boto3, 'resource')
    def test_connection_dequeue_message_with_wait_time(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        with mock.patch.object(SQSQueueMock, 'receive_messages') as receive_messages_mock:
            receive_messages_mock.return_value = [SQSMessageMock()]

            sqs_connector.dequeue(QUEUE_NAME, wait_time=10)

            receive_messages_mock.assert_called_with(
                AttributeNames=['All'],
                MaxNumberOfMessages=1,
                WaitTimeSeconds=10
            )

    @mock.patch.object(boto3, 'resource')
    def test_connection_dequeue_message_non_blocking_mode(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        with mock.patch.object(SQSQueueMock, 'receive_messages') as receive_messages_mock:
            receive_messages_mock.return_value = []

            payload = sqs_connector.dequeue(QUEUE_NAME, wait_time=0)

            assert payload is None

    @mock.patch.object(boto3, 'resource')
    def test_connection_delete_message_goes_ok(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        with mock.patch.object(SQSQueueMock, 'delete_messages') as delete_messages_mock:

            sqs_connector.delete(QUEUE_NAME, message_id=1)

            delete_messages_mock.assert_called_with(
                Entries=[
                    {
                        'Id': '1',
                        'ReceiptHandle': 1
                    }
                ]
            )

    @mock.patch.object(boto3, 'resource')
    def test_connection_delete_message_fails_if_no_queue_found(self, sqs_mock):
        sqs_mock.return_value = SQSMock(raise_queue_not_found=True)
        sqs_connector = self.create_sqs_connector()

        pytest.raises(QueueDoesNotExist, sqs_connector.delete, queue_name=QUEUE_NAME, message_id=1)

    @mock.patch.object(boto3, 'resource')
    def test_connection_retry_message_goes_ok(self, sqs_mock):
        sqs_mock.return_value = SQSMock()
        sqs_connector = self.create_sqs_connector()

        with mock.patch.object(SQSQueueMock, 'change_message_visibility_batch') as change_mock:

            sqs_connector.set_retry_time(QUEUE_NAME, message_id=5, delay=1)

            change_mock.assert_called_with(
                Entries=[
                    {
                        'Id': '1',
                        'ReceiptHandle': 5,
                        'VisibilityTimeout': 1
                    }
                ]
            )

            sqs_connector.set_retry_time(QUEUE_NAME, message_id=5, delay=None)

            change_mock.assert_called_with(
                Entries=[
                    {
                        'Id': '1',
                        'ReceiptHandle': 5,
                        'VisibilityTimeout': 0
                    }
                ]
            )

    @mock.patch.object(boto3, 'resource')
    def test_connection_retry_message_fails_if_no_queue_found(self, sqs_mock):
        sqs_mock.return_value = SQSMock(raise_queue_not_found=True)
        sqs_connector = self.create_sqs_connector()

        pytest.raises(
            QueueDoesNotExist,
            sqs_connector.set_retry_time,
            queue_name=QUEUE_NAME,
            message_id=1,
            delay=1
        )



from abc import ABCMeta, abstractmethod
from six import add_metaclass

from .exceptions import RetryException


@add_metaclass(ABCMeta)
class Job(object):
    """
    A simple queue job.

    Class attributes:
        * name: Name of the job. This name must be as unique as possible. Workers will decide
          which code to execute based on this name.
        * default_queue_name: Queue name where the job will be executed if one is not specified
          when adding the job to the broker.
        * abstract: Jobs marked as abstract will not be included by workers.
        * retry_time: Time (in seconds) used to define how long the message will be locked
          until other worker could retry the job. If None, it will use the queue's default value.

    Object attributes:
        * id: Unique ID of the job.
        * broker_id: Unique ID of the job given by the broker.
        * queue_name: Queue where the job has been added.
        * retries: How many retries has been done by the job.
        * created_on: When the job was enqueued for the first time.
    """
    name = None
    default_queue_name = 'sqjobs'
    abstract = False
    retry_time = None

    def __init__(self):
        self.id = None
        self.broker_id = None
        self.queue_name = None
        self.retries = 0
        self.created_on = None

    def __repr__(self):
        return '{0}()'.format(type(self).__name__)

    def pre_run(self, *args, **kwargs):
        pass

    @abstractmethod
    def run(self, *args, **kwargs):
        raise NotImplementedError

    def post_run(self, *args, **kwargs):
        pass

    def execute(self, *args, **kwargs):
        self.pre_run(*args, **kwargs)
        self.result = self.run(*args, **kwargs)
        self.post_run(*args, **kwargs)

        return self.result

    def on_success(self):
        pass

    def on_failure(self):
        pass

    def on_retry(self):
        pass

    def retry(self):
        raise RetryException

    def next_retry_time(self):
        return self.retry_time

    @classmethod
    def _task_name(cls):
        return cls.name or cls.__name__


class JobResult(object):
    """
    Result of adding a job to a broker.

    Object attributes:
        * job_id: Unique ID of the job.
        * broker_id: Unique ID of the job given by the broker.
        * result: Result of the execution (used by Eager mode).
    """

    def __init__(self):
        self.job_id = None
        self.broker_id = None
        self.result = None



from sqjobs import metadata
from sqjobs.brokers.eager import Eager
from sqjobs.brokers.standard import Standard
from sqjobs.brokers.multiqueue import MultiQueue
from sqjobs.connectors.sqs import SQS
from sqjobs.utils import (
    create_sqs_broker, create_sqs_worker, get_jobs_from_module,
    create_eager_broker, create_multiqueue_sqs_worker
)
from sqjobs.worker import Worker
from .fixtures import Adder, AbstractAdder


class TestBuilders(object):

    def test_eager_broker_builder(self):
        broker = create_eager_broker()
        assert isinstance(broker, Eager)

    def test_broker_builder(self):
        broker = create_sqs_broker('access', 'secret')

        assert isinstance(broker, Standard)
        assert isinstance(broker.connector, SQS)

        assert broker.connector.access_key == 'access'
        assert broker.connector.secret_key == 'secret'
        assert broker.connector.region_name == 'us-west-1'

    def test_broker_builder_with_custom_zone(self):
        broker = create_sqs_broker('access', 'secret', 'eu-west-1')
        assert broker.connector.region_name == 'eu-west-1'

    def test_worker_builder(self):
        worker = create_sqs_worker('queue_name', 'access', 'secret')

        assert isinstance(worker, Worker)
        assert isinstance(worker.broker, Standard)
        assert isinstance(worker.broker.connector, SQS)

        assert worker.queue_name == 'queue_name'
        assert worker.broker.connector.access_key == 'access'
        assert worker.broker.connector.secret_key == 'secret'
        assert worker.broker.connector.region_name == 'us-west-1'

    def test_multiqueue_worker_builder(self):
        worker = create_multiqueue_sqs_worker(['queue_name'], 'access', 'secret')

        assert isinstance(worker, Worker)
        assert isinstance(worker.broker, MultiQueue)
        assert isinstance(worker.broker.connector, SQS)

        assert worker.queue_name == ['queue_name']
        assert worker.broker.connector.access_key == 'access'
        assert worker.broker.connector.secret_key == 'secret'
        assert worker.broker.connector.region_name == 'us-west-1'

    def test_jobs_extracter(self):
        jobs = get_jobs_from_module('sqjobs.tests.fixtures')

        assert len(jobs) is 6
        assert Adder in jobs
        assert AbstractAdder not in jobs

    def test_invalid_module(self):
        jobs = get_jobs_from_module('invalid.module')
        assert jobs == []

    def test_min_metadata_info(self):
        assert metadata.__uname__ is not None
        assert metadata.__version__ is not None
        assert metadata.__license__ is not None



import importlib
import inspect

from .job import Job
from .brokers.standard import Standard
from .brokers.multiqueue import MultiQueue
from .brokers.eager import Eager
from .connectors.sqs import SQS
from .worker import Worker

import logging
logger = logging.getLogger('sqjobs')


def create_eager_broker():
    return Eager()


def create_sqs_broker(access_key, secret_key, region_name='us-west-1', endpoint_url=None):
    sqs = SQS(
        access_key=access_key,
        secret_key=secret_key,
        region_name=region_name,
        endpoint_url=endpoint_url,
    )

    return Standard(sqs)


def create_sqs_worker(queue_name, access_key, secret_key, region_name='us-west-1', endpoint_url=None):
    broker = create_sqs_broker(access_key, secret_key, region_name, endpoint_url)
    return Worker(broker, queue_name)


def create_multiqueue_sqs_worker(queue_names, access_key, secret_key, region_name='us-west-1', endpoint_url=None):
    sqs = SQS(
        access_key=access_key,
        secret_key=secret_key,
        region_name=region_name,
        endpoint_url=endpoint_url,
    )
    return Worker(MultiQueue(sqs), queue_names)


def get_jobs_from_module(module_name):
    jobs = []

    try:
        module = importlib.import_module(module_name)
    except ImportError:
        return jobs

    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, Job) and not obj.abstract and obj is not Job:
            jobs.append(obj)

    return jobs



import imp

from setuptools import setup, find_packages

metadata = imp.load_source('metadata', 'sqjobs/metadata.py')

setup(
    name=metadata.__uname__,
    version=metadata.__version__,
    url=metadata.__url__,
    license=metadata.__license__,
    author=metadata.__author__,
    author_email=metadata.__email__,
    description=metadata.__long_name__,
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    entry_points={
        'console_scripts': ['sqjobs = sqjobs.cli:main']
    },
    install_requires=[req.strip() for req in open('requirements/base.txt').readlines()],
    classifiers=[
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Systems Administration',
    ]
)



from django.test import TestCase
from django.core.management import call_command
from sqjobs.brokers.standard import Standard
from sqjobs.brokers.eager import Eager
from sqjobs.contrib.django.djsqjobs.models import PeriodicJob
from sqjobs.contrib.django.djsqjobs.models import JobStatus


from ...fixtures import Adder, Divider
from .django_fixtures import ResultDivider

from datetime import datetime
import pytz

from mock.mock import patch


class DjangoTestCase(TestCase):
    fixtures = ['sqjobs/tests/djsqjobs_tests-fixture.json']

    def setUp(self):
        call_command('loaddata', self.fixtures[0], verbosity=0)


class ResultJobsTests(DjangoTestCase):

    def test_result_jobs(self):
        result_divider = ResultDivider()
        result_divider.id = '1234'
        self.assertFalse(result_divider.properly_setup)
        result_divider.pre_run()
        self.assertEqual(repr(result_divider.job_status),
            '<JobStatus (ResultDivider()): 1234 status->PENDING>'
        )
        self.assertEqual(
            result_divider.job_status.to_dict(),
            {
                'status': 'PENDING',
                'job_id': '1234',
                'date_done': None,
                'traceback': None,
                'result': None,
                'date_started': None,
                'job_name': 'ResultDivider()'
            }
        )
        self.assertTrue(result_divider.properly_setup)
        self.assertEqual(result_divider.job_status.status, JobStatus.PENDING)
        result_divider.run(3, 1)
        self.assertTrue(result_divider.properly_setup)
        result_divider.on_success()
        self.assertEqual(result_divider.job_status.status, JobStatus.SUCCESS)
        result_divider.on_failure()
        self.assertEqual(result_divider.job_status.status, JobStatus.FAILURE)
        self.assertEqual(result_divider.err, 'ZeroDivisionError')
        result_divider.post_run()
        self.assertLessEqual(result_divider.job_status.date_done, datetime.now())
        self.assertEqual(result_divider.job_status.result, '3')



from abc import ABCMeta, abstractmethod
from six import add_metaclass
from uuid import uuid4


@add_metaclass(ABCMeta)
class Broker(object):

    @abstractmethod
    def add_job(self, job_class, *args, **kwargs):
        """
        Add a job to the broker

        :param job_class: python class of the payload job
        :param args: arguments to execute the job
        :param kwargs: keyword arguments to execute the job
        """
        raise NotImplementedError

    def gen_job_id(self):
        """
        Generate a new unique job ID
        """
        return str(uuid4())

    def serialize_job(self, job_name, job_id, args, kwargs):
        """
        Serialize a job into a string to be sent to the broker

        :param job_name: the name of the job
        :param job_id: the ID of the job
        :param args: arguments of the job
        :param kwargs: keyword arguments of the job
        """
        return self.connector.serialize_job(job_name, job_id, args, kwargs)

    def unserialize_job(self, job_class, payload):
        """
        Build a job given a payload returned from the broker

        :param job_class: python class of the payload job
        :param payload: python dict with the job arguments
        """
        return self.connector.unserialize_job(job_class, payload)



import base64
import json
import sys
import os
from datetime import date, datetime
from pytz import timezone

import boto3
import botocore

from .base import Connector
from ..exceptions import QueueDoesNotExist

import logging
logger = logging.getLogger('sqjobs.sqs')

is_pypy = '__pypy__' in sys.builtin_module_names


class SQS(Connector):
    """
    Manages a single connection to SQS
    """

    def __init__(self, access_key, secret_key, region_name='us-east-1', endpoint_url=None):
        """
        Creates a new SQS object

        :param access_key: access key with write access to AWS SQS
        :param secret_key: secret key with write access to AWS SQS
        :param region_name: a region name, like 'us-east-1'
        :param endpoint_url: URL to use a custom region
        """
        self.access_key = access_key
        self.secret_key = secret_key
        self.region_name = region_name
        self.endpoint_url = endpoint_url

        self._cached_connection = None

    def __repr__(self):
        return 'SQS("{ak}", "{sk}", region_name="{region_name}")'.format(
            ak=self.access_key,
            sk="%s******%s" % (self.secret_key[0:6], self.secret_key[-4:]),
            region_name=self.region_name,
        )

    @property
    def connection(self):
        """
        Creates (and saves in a cache) a connection to SQS
        """
        if self._cached_connection is None:
            config = {
                'service_name': 'sqs',
                'region_name': self.region_name,
                'aws_access_key_id': self.access_key,
                'aws_secret_access_key': self.secret_key,
            }

            if self.endpoint_url:
                config['endpoint_url'] = self.endpoint_url

            self._cached_connection = boto3.resource(**config)
            logger.debug('Created a new connection to SQS')

        return self._cached_connection

    def enqueue(self, queue_name, payload):
        message = SQSMessage.encode(payload)
        queue = self._get_queue(queue_name)

        if not queue:
            raise QueueDoesNotExist('The queue %s does not exist' % queue_name)

        queue.send_message(MessageBody=message)
        logger.info('Sent new message to %s', queue_name)

    def dequeue(self, queue_name, wait_time=20):
        queue = self._get_queue(queue_name)
        messages = None

        if not queue:
            raise QueueDoesNotExist('The queue %s does not exist' % queue_name)

        while not messages:
            messages = queue.receive_messages(
                MaxNumberOfMessages=1,
                WaitTimeSeconds=wait_time,
                AttributeNames=['All'],
            )

            if not messages:
                logger.debug('No message retrieved from %s', queue_name)

                if wait_time == 0:
                    return None  # Non-blocking mode

        logger.info('New message retrieved from %s', queue_name)
        payload = SQSMessage.decode(messages[0], queue_name)

        return payload

    def delete(self, queue_name, message_id):
        queue = self._get_queue(queue_name)

        if not queue:
            raise QueueDoesNotExist('The queue %s does not exist' % queue_name)

        queue.delete_messages(Entries=[{
            'Id': '1',
            'ReceiptHandle': message_id
        }])

        logger.info('Deleted message from queue %s', queue_name)

    def set_retry_time(self, queue_name, message_id, delay):
        queue = self._get_queue(queue_name)

        if not queue:
            raise QueueDoesNotExist('The queue %s does not exist' % queue_name)

        queue.change_message_visibility_batch(Entries=[{
            'Id': '1',
            'ReceiptHandle': message_id,
            'VisibilityTimeout': delay or 0
        }])

        logger.info('Changed retry time of a message from queue %s', queue_name)

    def serialize_job(self, job_name, job_id, args, kwargs):
        return {
            'id': job_id,
            'name': job_name,
            'args': args,
            'kwargs': kwargs
        }

    def unserialize_job(self, job_class, payload):
        job = job_class()

        job.id = payload['id']
        job.queue_name = payload['_metadata']['queue_name']
        job.broker_id = payload['_metadata']['id']
        job.retries = payload['_metadata']['retries']
        job.created_on = payload['_metadata']['created_on']
        args = payload['args'] or []
        kwargs = payload['kwargs'] or {}

        return job, args, kwargs

    def _get_queue(self, name):
        try:
            environment = os.environ.get('ENVIRONMENT', False)
            aws_queues_enabled = os.environ.get('AWS_SAM_QUEUES', False)
            if environment and aws_queues_enabled:
                # running in the AWS environments
                name = f'{environment}-{name}'
            return self.connection.get_queue_by_name(QueueName=name)
        except botocore.exceptions.ClientError:
            return None


class SQSMessage(object):

    @staticmethod
    def encode(payload):
        payload_str = json.dumps(payload, default=SQSMessage.json_formatter)
        payload_encoded = base64.b64encode(payload_str.encode('utf-8'))
        return payload_encoded.decode('utf-8')

    @staticmethod
    def decode(message, queue_name):
        body = message.body

        if is_pypy:
            body = body.encode("utf-8")

        payload_decoded = base64.b64decode(body)
        payload = json.loads(payload_decoded.decode("utf-8"))

        retries = int(message.attributes['ApproximateReceiveCount'])
        created_on = int(message.attributes['SentTimestamp'])

        payload['_metadata'] = {
            'id': message.receipt_handle,
            'retries': retries,
            'created_on': datetime.fromtimestamp(created_on / 1000, tz=timezone('UTC')),
            'queue_name': queue_name
        }

        logging.debug('Message payload: %s', str(payload))

        return payload

    @staticmethod
    def json_formatter(obj):
        if isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, date):
            return obj.strftime('%Y-%m-%d')

        return None
